// Ocean Shader by afl_ext
// MIT License
// https://www.shadertoy.com/view/MdXyzX
// Mostly copied verbatim with minor changes to formatting for readibility
#version 120

uniform float time;     // Time in seconds
const float camy=1.5;   // Camera elevation
const float depth=1.0;  // Wave depth
const float drag=0.28;  // How much waves pull on the water

#define ITERATIONS_RAYMARCH 12 // waves iterations of raymarching
#define ITERATIONS_NORMAL   40 // waves iterations when calculating normals

// Calculates wave value and its derivative, 
// for the wave direction, position in space, wave frequency and time
vec2 WaveDx(vec2 position,vec2 direction,float frequency,float timeshift)
{
   float x    = dot(direction,position)*frequency + timeshift;
   float wave = exp(sin(x)-1);
   float dwdx = -wave*cos(x);
   return vec2(wave,dwdx);
}

// Calculates waves by summing octaves of various waves with various parameters
float CalcWave(vec2 position,int iterations)
{
   float iter = 0; // this will help generating well distributed wave directions
   float frequency = 1; // frequency of the wave, this will change every iteration
   float timeMultiplier = 2; // time multiplier for the wave, this will change every iteration
   float weight = 1;// weight in final sum for the wave, this will change every iteration
   float sumOfValues = 0; // will store final sum of values
   float sumOfWeights = 0; // will store final sum of weights
   for (int i=0;i<iterations;i++)
   {
      // generate some wave direction that looks kind of random
      vec2 p = vec2(sin(iter),cos(iter));
      // calculate wave data
      vec2 res = WaveDx(position, p, frequency, time * timeMultiplier);
      // shift position around according to wave drag and derivative of the wave
      position += p * res.y * weight * drag;
      // add the results to sums
      sumOfValues += res.x * weight;
      sumOfWeights += weight;
      // modify next octave parameters
      weight *= 0.82;
      frequency *= 1.18;
      timeMultiplier *= 1.07;
      // add some kind of random value to make next wave look random too
      iter += 1232.399963;
   }
   // calculate and return
   return sumOfValues/sumOfWeights;
}

// Raymarches the ray from top water layer boundary to low water layer boundary
// Returns distance to intersection
float RayMarchWater(vec3 camera,vec3 start,vec3 end,float depth)
{
   vec3 pos = start;
   vec3 dir = normalize(end-start);
   for(int i=0;i<64;i++)
   {
      // The height is from 0 to -depth
      float height = CalcWave(pos.xz,ITERATIONS_RAYMARCH)*depth - depth;
      // If the waves height almost nearly matches the ray height, assume its a hit and return the hit distance
      if (height+0.01 > pos.y) return distance(pos,camera);
      // iterate forwards according to the height mismatch
      pos += dir*(pos.y-height);
   }
   // if hit was not registered, just assume hit the top layer, 
   // this makes the raymarching faster and looks better at higher distances
   return distance(start,camera);
}

// Calculate normal at point by calculating the height at the pos and 2 additional points very close to pos
vec3 Normal(vec2 pos,float e,float depth)
{
   vec2 ex = vec2(e,0);
   vec3 a = vec3(pos.x  ,CalcWave(pos.xy      ,ITERATIONS_NORMAL)*depth,pos.y);
   vec3 b = vec3(pos.x-e,CalcWave(pos.xy-ex.xy,ITERATIONS_NORMAL)*depth,pos.y);
   vec3 c = vec3(pos.x  ,CalcWave(pos.xy+ex.yx,ITERATIONS_NORMAL)*depth,pos.y+e);
   return normalize(cross(a-b,a-c));
}

//
//  Rotation matrix
//    angle in radions
//    x,y,z must be pre-normalized
//
mat3 RotMat(float th,float x,float y,float z)
{
   float s = sin(th);
   float c = cos(th);
   return mat3((1-c)*x*x+c   , (1-c)*x*y+z*s , (1-c)*z*x-y*s ,
               (1-c)*x*y-z*s , (1-c)*y*y+c   , (1-c)*y*z+x*s ,
               (1-c)*z*x+y*s , (1-c)*y*z-x*s , (1-c)*z*z+c   );
}

// Ray-Plane intersection distance
float PlaneDist(vec3 origin,vec3 direction,float z)
{ 
   vec3 normal = vec3(0,1,0);
   return clamp(dot(vec3(0,z,0)-origin,normal)/dot(direction,normal), -1, 9991999); 
}

// Very barebones but fast atmosphere approximation
vec3 CheapSky(vec3 raydir,vec3 sundir)
{
   sundir.y = max(-0.07,sundir.y);
   float raymul  = 1/(raydir.y+0.1);
   float raysun  = pow(abs(dot(sundir,raydir)),2);
   vec3 skyblue  = vec3(5.5,13.0,22.4)/22.4;
   vec3 suncolor = mix(vec3(1), max(vec3(0),1-skyblue), 1/(11*sundir.y+1));
   vec3 skycolor = max(vec3(0), skyblue*(suncolor - 0.0448*(raymul-6*sundir.y*sundir.y)));
   skycolor *= 0.24*(raysun+1)*(1+pow(1-raydir.y,3));
   suncolor *= 0.2*pow(max(0,dot(sundir,raydir)),8);
   return raymul*(skycolor + suncolor);
} 

//  Get color of sky with sun
vec3 SkyColor(vec3 dir)
{
   //  Calculate where the sun should be, it will be moving around the sky
   vec3 sundir = normalize(vec3(sin(0.1*time),1,cos(0.1*time)));
   //  Sky color
   vec3 sky = 0.5*CheapSky(dir,sundir);
   //  Sun color
   float sun = 210*pow(max(0,dot(dir,sundir)),720);
   return sky + sun;
}

// Great tonemapping function from other afl_ext shader: https://www.shadertoy.com/view/XsGfWV
vec4 aces_tonemap(vec3 color)
{
   mat3 m1 = mat3(
     0.59719, 0.07600, 0.02840,
     0.35458, 0.90834, 0.13383,
     0.04823, 0.01566, 0.83777
   );
   mat3 m2 = mat3(
     1.60475, -0.10208, -0.00327,
     -0.53108,  1.10813, -0.07276,
     -0.07367, -0.00605,  1.07602
   );
   vec3 v = 2*m1*color;  
   vec3 a = v*v+0.024578*v-0.000090537;
   vec3 b = 0.983729*v*v+0.4329510*v+0.238081;
   v = m2*(a/b);
   return vec4(pow(clamp(v,0,1),vec3(1/2.2)),1);
}

// Main
void main()
{
   // Initialize ray direction for every pixel based on texture coordinates
   vec3 dir = normalize(vec3(gl_TexCoord[0].xy-0.9,1.5));
   // If dir.y is positive, render sky
   if(dir.y >= 0)
   {
      gl_FragColor = aces_tonemap(SkyColor(dir));
   }
   // If dir.y is negative, render water
   else
   {
      // Ray origin, moving around
      vec3 origin = vec3(time,camy,time);

      // Calculate where ray hits high (0) horizontal plane
      float dist = PlaneDist(origin, dir, 0);
      vec3 hiHitPos = origin + dist*dir;
      // Calculate where ray hits low (-depth) horizontal plane
      dist = PlaneDist(origin, dir, -depth);
      // Calculate hit locations for high and low planes
      vec3 loHitPos = origin + dist*dir;

      // Raymatch water.  Will be between high and low planes
      dist = RayMarchWater(origin, hiHitPos, loHitPos, depth);
      vec3 waterHitPos = origin + dist*dir;

      // Calculate normal at the hit position
      vec3 N = Normal(waterHitPos.xz, 0.01, depth);

      // Smooth the normal with distance to avoid disturbing high frequency noise
      N = mix(N, vec3(0,1,0) , 0.8*min(1,1.1*sqrt(0.01*dist)));

      // Reflect the ray and make sure it bounces up
      vec3 R = normalize(reflect(dir,N));
      R.y = abs(R.y);
      
      // Calculate the reflection and approximate subsurface scattering
      vec3 reflection = SkyColor(R);
      vec3 scattering = vec3(0.0293,0.0698,0.1717) * 0.1*((waterHitPos.y+depth)/depth+0.2);

      // Calculate fresnel coefficient
      float fresnel = 0.96*pow(1-max(0,dot(-N,dir)),5)+0.04;
      // Combined reflection and scattering
      gl_FragColor = aces_tonemap(fresnel*reflection + (1-fresnel)*scattering);
   }
}
